# This file exists because I wanted to keep ppp_stdlib.py and ppp_interpreter.py seperate but they both rely on this one class.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, Dict, List as List_, Tuple as Tuple_, Union as Union_

from ppp_ast import Statement
from ppp_types import ArrayType, EnumType, FunctionType, ListType, ReturnType, StructType, TupleType, Type, Int as IntType, Str as StrType, Bool as BoolType, Void as VoidType, TypeType

class Object(ABC):
	@abstractmethod
	def get_type(self) -> Type: ...

	def hash(self) -> 'Hashable':
		assert False, f"{self.get_type().represent()} cannot be hashed."

@dataclass
class Int(Object):
	num: int

	def get_type(self) -> Type: return IntType

@dataclass
class Str(Object):
	str: str

	def get_type(self) -> Type: return StrType

	def hash(self) -> 'HStr':
		return HStr(self.str)

@dataclass
class Bool(Object):
	value: bool

	def get_type(self) -> Type: return BoolType

@dataclass
class Void_(Object):
	def get_type(self) -> Type: return VoidType
Void = Void_()

@dataclass
class TypeObject(Object):
	type: Type

	def get_type(self) -> Type: return TypeType

@dataclass
class Tuple(Object):
	type: TupleType
	tuple: Tuple_[Object, ...]

	def get_type(self) -> Type: return self.type

@dataclass
class List(Object):
	type: ListType
	list: List_[Object]

	def get_type(self) -> Type: return self.type

@dataclass
class Array(Object):
	type: ArrayType
	array: List_[Object]

	def get_type(self) -> Type: return self.type

@dataclass
class Function(Object):
	type: FunctionType
	function: Tuple_[str, List_[Tuple_[str, Type]], Type, Statement, Callable[..., Object]]

	def get_type(self) -> Type: return self.type

@dataclass
class Return(Object):
	type: ReturnType
	value: Object

	def get_type(self) -> Type: return self.type

@dataclass
class EnumValue(Object):
	type: EnumType
	name: str
	values: List_[Object]

	def get_type(self) -> Type: return self.type

	def hash(self) -> 'HEnumValue':
		return HEnumValue(self.type, self.name, [value.hash() for value in self.values])

@dataclass
class Struct(Object):
	type: StructType
	fields: Dict[str, Object]

	def get_type(self) -> Type: return self.type

class Hashable(ABC):
	@abstractmethod
	def __hash__(self) -> int: ...

	@abstractmethod
	def get_object(self) -> Object: ...

@dataclass
class HInt(Hashable):
	num: int

@dataclass
class HStr(Hashable):
	str: str

	def __hash__(self) -> int:
		return hash(('object', 'str', self.str))

	def get_object(self) -> Object:
		return Str(self.str)

@dataclass
class HEnumValue(Hashable):
	type: EnumType
	name: str
	values: List_[Hashable]

	def __hash__(self) -> int:
		return hash(('object', 'enum', self.type, self.name, tuple(self.values)))

	def get_object(self) -> Object:
		return EnumValue(self.type, self.name, [value.get_object() for value in self.values])
